params = {
    "capitalists": (200, 700),
    "cpf_agents": (2, 20),
    "csf_agents": (2, 20),
    "green_energy_owners": (1, 200),
    "brown_energy_owners": (1, 200),
    "unemploymentDole": (100, 2000),
    "tauDesiredConsumption": (0.1, 0.5),
    "ADAPTIVE_PRICE_CHANGE": (0.001, 0.5),
    "ADAPTIVE_PRICE_CHANGE_K": (0.001, 0.5),
    "ADAPTIVE_PRICE_CHANGE_ENERGY": (0.001, 0.5),
    "fossilFuelPriceGrowthRate": (0.001, 0.01),
    "rhoForecast": (0.1, 0.9),
    "rhoForecastEnergy": (0.1, 0.9),
    "rhoForecastCapital": (0.1, 0.9),
    "wageAdjustmentRate": (0.1, 0.9),
    "shareOfWageInConsumption": (0.1, 0.8),
    "shareOfIncomeInConsumption": (0.1, 0.8),
    "shareOfWealthInConsumption": (0.0, 0.8),  # 0.2
    "ownerProportionFromProfits": (0.1, 0.5),
    "repayProportionFromProfits": (0.01, 0.1),
    "wageDividendFromProfits": (0.1, 0.3),
    "capital_multiplier": (1.1, 3),
    "MARKUP_ADJUSTMENT": (0.1, 0.9),
    "rho_labour": (0.1, 0.9),
    "rho_energy": (0.1, 0.9),
    "rho_labour_K": (0.1, 0.9),
    "rho_energy_K": (0.1, 0.9),
}



parameters = {
    "settings": "BAU",  # BAU, CT or CTR{a,b,c,d}
    "covid_settings": None,  # "BAU", "DIST", "LOCK", "VAX"
    "verboseFlag": False,
    "energySectorFlag": True,
    "climateModuleFlag": False,
    # Agents count (should be fixed)
    "c_agents": 5000,
    "capitalists": 150,
    "green_energy_owners": 12,
    "brown_energy_owners": 3,
    "b_agents": 1,
    "csf_agents": 6,
    "cpf_agents": 2,
    "g_agents": 1,
    # Time
    "start_date": "1980-01-01",
    "covid_start_date": "2020-01-01",
    "steps": 16435,  # + 365*2, # 8401, #4383 #36525
    "seed": 69,
    "capital_length": 20,
    # Consumer parameter
    "initialProbabilityToBeEmployed": 1,
    "unemploymentDole": 196.289062,  # (0, 1800)
    "minimumWage": 1800,
    "owner_endownment": 6241.40625,  # (2400, 20000)
    "worker_endownment": 2400,
    "wageAdjustmentRate": 0.001,  # in wage adjustment equation for C-firms #(0, 0.01)
    "subsistenceLevelOfConsumption": 49.5507812,  # (10, 50)
    "worker_additional_consumption": 5.09130859,  # (1, 50)int
    "owner_additional_consumption": 61,  # (50, 200)int
    "sick_reduction": 12,
    "consumption_growth": 0.00162,  # (0, 0.01)
    "consumption_var": 0.0719,  # (0, 0.1)
    "taxRate": 0.1,
    "incomeTaxRate": 0.1,
    "co2_price": 25,
    "co2_tax": 0.025,
    "ownerProportionFromProfits": 0.6558,  # (0, 1)
    "energyOwnerProportionFromProfits": 0.7944,  # (0, 1)
    "debtCancelledOnfinancialDifficulty": False,
    "bankResInit": 7.241e9,  # (0, 1e10)
    "bankIL": 0.1205,  # (0.0073, 0.2)
    "bankID": 0.0073,
    "bankICB": 0.002,
    "bankICB_P": 0.0075,
    "gammaRRR": 0.1,
    "iB": 0.1,
    "greenLoanRepayPeriod": 36,
    "brownLoanRepayPeriod": 36,
    "defaultProbAlpha": 1,
    # "targetInflation": 0.02,
    # "bankCredibility": 0.5,
    # "bankEquityEta": 0.08,
    # "inflationRateSensitivity": 0.1,
    "depreciationRate": 0.2,  # (0, 0.5)
    "forecast_discount_factor": 0.9067,  # (0, 1)
    "ETA_PRODUCTION_CAPITAL": (
        (2.15 - 1.0) / 2.15
    ),  # parameter regulating capital production, (calibrate the x inside of (x - 1)/x) #(1, 10)
    "ETA_PRODUCTION": (
        (2.0 - 1.0) / 2.0
    ),  # parameter regulating production function == might be the same value as the rho defined above #(1, 10)
    "mark_up_factor": 0.522,  # (0, 1)
    "mark_up_alpha": 0.556,  # (0.25, 1)
    "mark_up_beta": 0.444,  # (0, 0.75) #mark_up_alpha + mark_up_beta = 1
    # "capital_multiplier": 2.66,
    "capital_growth_rate": 0.0232,  # (0, 0.2)
    "mark_up_adjustment": 0.01,
    # Firms production parameter (fixed value)
    "rho_labour": 80,  # (10, 60)
    "rho_energy": 100,  # (60, 200)
    "rho_labour_K": 60,  # (30, 80)
    "rho_energy_K": 120,  # (80, 500)
    # beware these are the alphas of the production fucntion! of the C-firm
    "beta_labour": 0.25,
    "beta_capital": 0.4,
    "beta_energy": 0.35,
    # beware these are the alphas of the production fucntion! of the K-firm
    "beta_labour_K": 0.05,
    "beta_capital_K": 0.55,
    "beta_energy_K": 0.4,
    "reserve_ratio": 0.3,  # (0, 0.5)
    # For the energy sector
    "energy_price_growth": 0.0089,  # (0, 0.1)
    "base_green_energy_price": 60.254,  # (0, 200)
    "fossil_fuel_price": 1.514,  # (0, 100) (fossil_fuel_price < base_green_energy_price)
    "fossil_fuel_price_growth_rate": 0.00848,  # (0, 0.01)
    # None for turning off shock
    "climateShockMode": "None",
    # "climateShockMode": "Idiosyncratic",
    "climate_shock_start": 120,  # number of months until shock start to activate
    "climateWindSpeed": 0.7625,  # 0.9,
    "climateZetaGreen": 0.01,  # 0.01, # emission intensity green firms
    "climateZetaBrown": 0.5,  # 0.10, # emission intensity dirty firms
    "climateZetaAlpha": 0.2,
    "climateZetaBeta": 1e25,
    "CO2_offset": 8.03e9,
    "climateAlpha_conc": 0.00000015,  # 0.3, # parameter regulating emissions concentrations
    "climateBeta_conc": 0.7,  # 0.0005, # parameter regulating concentrations converge to pre-industrial levels
    "climateConc_t0": 1000,  # concentrations at the beginning of the simulation
    "climateConc_pre": 500,  # pre-industrial concentrations
    "climateGammaRF": 4.5,  # 1, # parameter regulating radiative force
    "climateSensitivity": 6.90625,  #
    "climateAlpha_phi": -42.7,  # in e-folding time equation phi = max(climateAlpha_phi + climateBeta_phiL*climateSensitivity + climateBeta_phiQ*climateSensitivity**2, 1)
    "climateBeta_phiL": 29.1,  # in e-folding time equation phi = max(climateAlpha_phi + climateBeta_phiL*climateSensitivity + climateBeta_phiQ*climateSensitivity**2, 1)
    "climateBeta_phiQ": 0.001,  # in e-folding time equation phi = max(climateAlpha_phi + climateBeta_phiL*climateSensitivity + climateBeta_phiQ*climateSensitivity**2, 1)
    "climateT0": 0,  # 1, # Temperature at time 0
    "climateAlpha_d": 0.05,  # Benchmark damage
    "climateSigma_d": 0.9,  # Storm sensitivity to atmospheric CO2 concentrations
    "climateEps_etd": -0.514,  # income elasticity of storm energy
    "climateGamma_etd": 1,  # a parameter for calculating Temperature damage on energy
    "climateBeta_d": 0.05,  # Benchmark mortality
    "climateEps_etm": -0.501,  # income elasticity of storm mortality and is calibrated following Toya and Skidmore 2007
    "climateGamnma_etm": 1,
    "psi_h": 0.1,
    "psi_f_g": 0.15,
    "psi_f_b": 0.25,
    "alpha_T": 1,
    "alpha_P": 1.5,
    "T_threshold": 10,
    "P_threshold": 15,
    ## Government
    "fiscal_time": 100,
    "csfirm_netWorth": 2750000,
    "cpfirm_netWorth": 82000000,
    "lumpSumState": True,
    "lumpSum": 100000,
    "csfirm_avgDeposit": 2000,
    "cpfirm_avgDeposit": 3000,
    "alpha_f": 0.1,
    "alpha_h": 0.1,
    ## Covid
    "production_cost": 500,
    "initialExposer": 1000,
    "dispersion_community": None,
    "overshoot_community": 1.2,
    "num_contacts_community": 20,
    "p_contact_community": 0.005,
    "dispersion_firms": None,
    "overshoot_firms": 1.2,
    "num_contacts_firms": 20,
    "p_contact_firms": 0.01,
    "p_mutation": 0.1,
    "T_susceptible_mild_mean": 8.5,  # 4.5
    "T_susceptible_mild_std": 1.5,
    "p_susceptible_mild_young": 0.9,
    "p_susceptible_mild_working": 0.9,
    "p_susceptible_mild_elderly": 0.9,
    "T_exposed_mild_mean": 3.1,  # 1.1
    "T_exposed_mild_std": 0.9,
    "p_exposed_mild_young": 0.7,
    "p_exposed_mild_working": 0.9,
    "p_exposed_mild_elderly": 0.9,
    "T_nonsym_recovered_mean": 8.0,
    "T_nonsym_recovered_std": 2.0,
    "T_mild_severe_mean": 10.6,  # 6.6
    "T_mild_severe_std": 4.9,
    "p_mild_severe_young": 0.005,
    "p_mild_severe_working": 0.7,
    "p_mild_severe_elderly": 0.25,
    "T_mild_recovered_mean": 8.0,
    "T_mild_recovered_std": 2.0,
    "T_severe_critical_mean": 1.5,
    "T_severe_critical_std": 2.0,
    "p_severe_critical_young": 0.001,
    "p_severe_critical_working": 0.01,
    "p_severe_critical_elderly": 0.8,
    "T_severe_recovered_mean": 18.1,
    "T_severe_recovered_std": 6.3,
    "T_critical_death_mean": 10.7,
    "T_critical_death_std": 4.8,
    "p_critical_death_young": 0.00001,
    "p_critical_death_working": 0.01,
    "p_critical_death_elderly": 0.2,
    "T_critical_recovered_mean": 18.1,
    "T_critical_recovered_std": 6.3,
    "p_recovered_immun_young": 0.99,
    "p_recovered_immun_working": 0.9,
    "p_recovered_immun_elderly": 0.7,
    ## Covid policies
    # Social Distancing
    "inf_threshold": 0.3,  # 0.6
    "p_sd": 0.1,  # (0,1)
    # Lock down
    "p_lockdown": 0.4,  # 0.5
    "duration_LD": 14,
    "num_C_firms_LD": 0.6,
    "num_K_firms_LD": 0.4,
    "pandemicWageTransfer": 3000,
    "lock_down_production_utilization": 0.6,
    # Vaccines
    "p_vax": 0.8,  # (0,8)
}



#!/usr/bin/env python3
"""
Basic functionality tests for CliMaPan-Lab.
"""

import os
import sys
import tempfile
import unittest

import numpy as np

# Add the climapan_lab package to the path
sys.path.insert(0, os.path.join(os.path.dirname(__file__), ".."))

from climapan_lab.base_params import economic_params
from climapan_lab.model import EconModel
from climapan_lab.src.consumers.Consumer import Consumer
from climapan_lab.src.firms.CapitalGoodsFirm import CapitalGoodsFirm
from climapan_lab.src.firms.ConsumerGoodsFirm import ConsumerGoodsFirm


class TestBasicFunctionality(unittest.TestCase):
    """Test basic functionality of the economic model."""

    def setUp(self):
        """Set up test fixtures."""
        self.params = economic_params.copy()
        # Use small numbers for testing
        self.params.update(
            {
                "c_agents": 10,
                "capitalists": 3,
                "csf_agents": 2,
                "cpf_agents": 2,  # Need at least 2 for energy type diversity
                "steps": 5,
                "verboseFlag": False,
                "climateModuleFlag": False,  # Disable for faster testing
                "covid_settings": None,  # Disable COVID for testing
            }
        )

    def test_model_creation(self):
        """Test that we can create a model instance."""
        model = EconModel(self.params)
        self.assertIsInstance(model, EconModel)
        # Model needs to be set up before agents are created
        model.setup()
        self.assertEqual(len(model.consumer_agents), self.params["c_agents"])

    def test_model_step(self):
        """Test that we can run a single step of the model."""
        model = EconModel(self.params)
        model.setup()
        # Test that step() executes without error
        try:
            model.step()
            step_completed = True
        except Exception:
            step_completed = False
        self.assertTrue(step_completed, "Model step should complete without error")

    def test_agent_creation(self):
        """Test that agents can be created with proper attributes."""
        model = EconModel(self.params)
        model.setup()

        # Test consumer creation
        consumer = model.consumer_agents[0]
        self.assertIsInstance(consumer, Consumer)
        self.assertTrue(hasattr(consumer, "deposit"))

        # Test firm creation
        if len(model.csfirm_agents) > 0:
            firm = model.csfirm_agents[0]
            self.assertIsInstance(firm, ConsumerGoodsFirm)
            self.assertTrue(hasattr(firm, "netWorth"))

    def test_parameters_validation(self):
        """Test that parameters can be loaded and models created."""
        # Test that model creation with valid parameters works
        model = EconModel(self.params)
        self.assertIsInstance(model, EconModel)

        # Test that model can access parameters
        self.assertTrue(hasattr(model, "p"))
        self.assertEqual(model.p.c_agents, self.params["c_agents"])

    def test_model_run_short(self):
        """Test that we can run the model for a few steps."""

    def test_model_run_short(self):
        """Test that we can run the model for a few steps."""
        model = EconModel(self.params)
        results = model.run()

        # Check that the model ran for the expected number of steps
        self.assertEqual(model.t, self.params["steps"])

        # Check that data collection works (ambr returns dict with 'model' DataFrame)
        # Legacy AgentPy checked model.output.variables
        # We now check the returned results
        if isinstance(results, dict):
            self.assertIn("model", results)
            self.assertFalse(results["model"].is_empty())
        else:
            # If wrapper is used or AgentPy compat matches
            self.assertTrue(hasattr(results, "variables"))
            self.assertTrue(hasattr(results.variables, "EconModel"))


class TestDataStructures(unittest.TestCase):
    """Test data structures and utilities."""

    def test_parameter_structure(self):
        """Test that economic parameters have expected structure."""
        params = economic_params

        # Check for essential parameters
        essential_keys = [
            "c_agents",
            "steps",
            "taxRate",
            "unemploymentDole",
            "subsistenceLevelOfConsumption",
        ]

        for key in essential_keys:
            self.assertIn(key, params, f"Missing essential parameter: {key}")

    def test_numpy_compatibility(self):
        """Test that the package works with numpy arrays."""
        test_array = np.array([1, 2, 3, 4, 5])
        self.assertEqual(len(test_array), 5)
        self.assertEqual(test_array.sum(), 15)


if __name__ == "__main__":
    unittest.main()



import copy
from collections import OrderedDict

import ambr as am
import numpy as np
import numpy.random as random
from scipy.optimize import minimize

from ..utils import days_in_month

# ============================================================================
#                           GoodsFirmBase
# ============================================================================
# Role:
#   Base class with finance, production cost accounting, labor management,
#   loans/repayment, carbon-tax handling, and shock/lockdown utilities.
#
# Lifecycle (per month; orchestrated by the model):
#   1) plan:    set planned_production (in subclass), call production_budgeting()
#   2) hire:    model assigns workers; calculate_all_wages() builds wage bill
#   3) produce: subclass computes actual_production via production_function()
#   4) sell:    model matches demand → set/updateSoldProducts(), setPrice()
#   5) settle:  compute costs, profits, update capital value, progressPayback()
#   6) tax/div: updateProfitsAfterTax(), ownerIncome handled at model level
#
# Key state:
#   capital, energy, workersList, wages, price, deposit, loans (loanList),
#   DTE (debt-to-equity), iL (loan rate), iF (financial fragility proxy),
#   average_production_cost, planned/actual production, carbonTax etc.
# ============================================================================


class GoodsFirmBase(am.Agent):
    """A GoodsFirmBase agent"""

    def setup(self):

        # ----------------------------------------
        # Core firm stocks & finance
        # ----------------------------------------
        self.workersList = []  # IDs of employed workers
        self.wages = {}  # id → wage mapping for this firm
        self.price = 0  # initial price of produced goods (set by model)
        self.capital = 0  # physical units of installed capital
        self.energy = 0  # energy input planned/used this step
        self.deposit = 0  # cash / liquid assets
        self.equity = 0
        self.net_profit = 0
        self.non_loan = 0  # written-off part after bankruptcy

        # ----------------------------------------
        # Technology & policy parameters (snapshot from self.p)
        # ----------------------------------------
        self.iL = self.p.bankICB  # base loan rate (bank interbank proxy)
        self.beta_capital = self.p.beta_capital
        self.beta_labour = self.p.beta_labour
        self.beta_energy = self.p.beta_energy
        self.eta_production = self.p.ETA_PRODUCTION
        self.rho_labour = self.p.rho_labour
        self.rho_energy = self.p.rho_energy
        self.capital_growth_rate = self.p.capital_growth_rate
        self.carbon_tax_state = self.p.settings.find("CT") != -1
        self.div_ratio = self.p.ownerProportionFromProfits
        self.capital_depreciation = self.p.depreciationRate
        self.wage_factor = 1 / (1 + self.p.wageAdjustmentRate)

        # ----------------------------------------
        # Transient/accounting variables
        # ----------------------------------------
        self.wage_bill = 0
        self.unitWageBill = self.p.unemploymentDole
        self.capital_tracking = []  # rolling buffer of investment values
        self.actual_production = 0  # 10
        self.planned_production = 0
        self.average_production_cost = 0
        self.labour_demand = 0
        self.aggregate_demand = 0
        self.priceList = [1]

        # Capital purchase & valuation
        self.capital_investment = 0  # value variable
        self.capital_increase = 0  # physical variable
        self.capital_price = 1
        self.capital_value = self.capital * self.capital_price  # value variable
        self.cost_of_capital = 0  # value variable - amortized cost in avg cost calc
        self.capital_purchase = 0  # physical variable
        self.capital_demand = 0  # physical variable
        self.capital_growth = 0  # physical variable

        # Energy mix
        self.useEnergy = None  # 'green' or 'brown'
        self.brown_firm = self.useEnergy == "brown"

        # Data writer for quick diagnostics
        self.firmDataWriter = []  # [sold, wage_bill, n_workers] each step

        # Banking / leverage
        self.loanObtained = 0
        self.loan_demand = 0
        self.loanList = [0, 0]  # outstanding principals by "vintage"
        self.loanContractRemainingTime = {}  # vintage → months remaining
        self.DTE = 0  # debt-to-equity ratio
        self.iF = 0  # interest paid on loans
        self.reserve_ratio = self.p.reserve_ratio

        # Other accounting & flags
        self.depositList = [0, 0]
        self.sale_record = 0
        self.profits = 0
        self.fix_cost = 0
        self.ownerIncome = 0
        self.netWorth = 0
        self.countConsumers = 0
        self.countWorkers = 0
        self.carbonTax = 0
        self.defaultProb = 0
        self.bankrupt = False
        self.lockdown = False
        self.lockdownList = []
        self.payback = 0
        self.profit_margin = 0

        # Shortcuts to agent lists
        self.consumersList = self.model.consumer_agents
        self.sickList = self.model.consumer_agents

        # Taxes & sales
        self.tax = 0
        self.soldProducts = 1

        # ----------------------------------------
        # Rebuild worker roster if reloaded mid-run
        # ----------------------------------------
        self.workersList = []
        [
            self.workersList.append(aConsumer.id)
            for aConsumer in self.model.aliveConsumers
            if aConsumer.consumerType == "workers"
            and aConsumer.getBelongToFirm() == self.id
            and aConsumer.getCovidStateAttr("state") != "dead"
        ]
        for aConsumer in self.model.aliveConsumers:
            if aConsumer.getIdentity() in self.workersList:
                self.wages[aConsumer.getIdentity()] = aConsumer.getWage()

    # ========================================
    # Bankruptcy reset (re-entry with new balance sheet)
    # ========================================
    def bankrupt_reset(self):
        """Reset firm state after bankruptcy"""
        print("firm get bankrupt!!!", self.id)
        self.netWorth = 0
        self.loanObtained = 0
        self.loanList = [0, 0]
        self.loanContractRemainingTime = {}
        self.payback = 0
        self.bankrupt = False
        self.DTE = 0
        self.deposit = 0
        self.depositList = [0, 0]
        self.fiscal = 0
        self.brown_firm = self.useEnergy == "brown"

    # ========================================
    # CES-like production function (unit-consistent)
    # ========================================
    def production_function(self, inputs):
        """Define consumer goods production function"""
        capital_input, labour_input, energy_input = inputs
        production = (
            self.beta_capital * (capital_input) ** self.eta_production
            + self.beta_labour * (labour_input * self.rho_labour) ** self.eta_production
            + self.beta_energy * (energy_input * self.rho_energy) ** self.eta_production
        ) ** (1 / self.eta_production)
        return production

    # ========================================
    # Financing the plan (loan demand, deposits, logging)
    # ========================================
    def production_budgeting(self):
        """Decide funding for planned production; compute loan demand."""
        # print("capital stock", self.capital, self.capital_increase, self.capital_growth)
        # checking if production can be financed, otherwise taking loan if possible
        # Need enough cash to cover variable cost; allow some minimum leverage
        self.loan_demand = np.max(
            [
                self.get_average_production_cost() * self.planned_production
                - self.deposit * (1 - self.reserve_ratio),
                self.get_average_production_cost() * self.planned_production * 0.2,
            ]
        )
        # self.loan_demand = np.max([self.get_average_production_cost() * self.planned_production, 0])
        if self.loan_demand > 0:
            # Bank module will respond via adjustAccordingToBankState
            pass
        else:
            self.loanList.append(0)

            data = [self.getSoldProducts(), self.wage_bill, len(self.workersList)]
            self.firmDataWriter.append(data)

    # ========================================
    # Average production cost decomposition (wage, energy, capital)
    # ========================================
    def calculate_average_production_cost(self):
        """Function to derive production cost"""
        ## Wage Component
        # Wage Component
        # Wage component (use model-wide mean wage among workers)
        self.calculate_all_wages()

        # Optimization: Avoid re-selecting workers from global list
        # consumersList is likely self.model.aliveConsumers (shared)
        # We can use the cached 'workers' subset if available, or vectorize.

        # Original: self.consumersList.select(consumerType == "workers").getWage()
        # This is slow O(N).
        # We know who workers are: self.model.aliveConsumers where consumerType=='workers'
        # Even better, we performed wage updates in calculate_all_wages on self.workersList (IDs).

        # If we need the GLOBAL mean wage of workers:

        consumers = self.consumersList
        # Vectorized access (AMBER or AgentPy list)
        # If consumers is an AgentList, it has attribute access.
        # wages = consumers.wage  <-- but we need to filter by type "workers"

        # Fast way:
        # wages = np.array(consumers.wage)
        # types = np.array(consumers.consumerType)
        # mean_wage = np.mean(wages[types == "workers"])

        # Even faster: The Bank or Model likely knows this?
        # But to be safe and local:

        # Note: 'consumersList' is passed to firm.
        # If we use list comprehension it is faster than .select()

        # However, calling this for EVERY firm is wasteful if it's the same global value.
        # But assuming we must keep it local:

        # Gather all wages of workers.
        # Since we just updated wages, we can iterate self.wages (dict of {id: wage})
        # IF self.wages contains ALL workers.
        # In `calculate_all_wages`, we iterate `consumersList`, filter by `workersList` (global active workers?),
        # and update `self.wages`.
        # So `self.wages` SHOULD contain the latest wages of relevant workers.
        # CHECK: Is self.wages specific to this firm or global? Base class has `self.wages = {}`.
        # Each firm calculates wages for... who?
        # In `calculate_all_wages`, it iterates `self.consumersList` (passed in Init).
        # If `consumersList` is ALL consumers, then every firm calculates wages for ALL consumers?
        # That would be redundant!
        # `GoodsFirmBase` calls `calculate_all_wages`... wait.
        # If every firm calls `calculate_all_wages` and it loops over ALL consumers, that is O(Firms * Consumers).
        # That explains why `calculate_all_wages` was slow!
        # I optimized the loop, but the architecture seems to have every firm updating every consumer??
        # If so, `mean_wage` is just `np.mean(list(self.wages.values()))`.

        if self.wages:
            mean_wage = np.mean(list(self.wages.values()))
        else:
            mean_wage = 0  # Should not happen if workers exist
        # it is probably best to just sum the wage instead of taking average and multiply with number of labour

        ## Energy Component
        # Energy component (price depends on current mix)
        energy_price = (
            self.model.brownEFirm[-1].getPrice()
            if self.brown_firm
            else self.model.greenEFirm[-1].getPrice()
        )
        energy_cost = self.get_energy() * energy_price
        # print("energy break down: ", self.id, self.getUseEnergy(), self.get_energy(), energy_price)

        ## Capital Component
        # Capital component (amortize recent investment over a rolling window)
        # Need to default value of past periods to 0
        if len(self.capital_tracking) > self.p.capital_length:
            self.capital_tracking.pop(0)
        self.capital_tracking.append(np.sum(self.capital_investment))

        cost_of_capital = np.sum(self.capital_tracking) / self.p.capital_length
        self.cost_of_capital = cost_of_capital
        ##Need to also store the price of capital together with the period so we can do multiplication

        # Unit cost and fixed cost (wage+capital, excluding variable energy)
        self.average_production_cost = (
            mean_wage * self.getNumberOfLabours() + energy_cost + cost_of_capital
        ) / (self.get_actual_production())
        # print(self.id, " cost of each type: ", self.getUseEnergy(), " wage ", mean_wage * self.getNumberOfLabours(),
        #      " energy ", energy_cost,
        #      " capital ", cost_of_capital, "-", self.capital_price,
        #      " production ", self.get_actual_production())

        self.fix_cost = mean_wage * self.getNumberOfLabours() + cost_of_capital

    # ========================================
    # Loan amortization & delinquency evolution (monthly)
    # ========================================
    def payLoan(self):
        """Amortize current loans with available payback amount"""
        if -self.payback > 0:
            payback = copy.copy(self.payback)
            for loan_id in [
                i for i in range(len(self.loanList)) if self.loanList[i] > 0
            ]:
                if payback == 0:
                    break
                if -payback <= self.loanList[loan_id]:
                    self.loanList[loan_id] += np.sum(payback)
                    payback = 0
                else:
                    self.loanList[loan_id] = 0
                    payback += np.sum(self.loanList[loan_id])
                # Contract time and default trigger
                if (
                    self.loanList[loan_id] > 0
                    and self.loanContractRemainingTime[loan_id] <= 0
                ):
                    self.bankrupt = True
                elif self.loanList[loan_id] > 0:
                    self.loanContractRemainingTime[loan_id] -= 1
                else:
                    self.loanContractRemainingTime.pop(loan_id)
        else:
            # No payment this step; still age contracts and check default
            for loan_id in [
                i for i in range(len(self.loanList)) if self.loanList[i] > 0
            ]:
                if (
                    self.loanList[loan_id] > 0
                    and self.loanContractRemainingTime[loan_id] <= 0
                ):
                    self.bankrupt = True
                elif self.loanList[loan_id] > 0:
                    self.loanContractRemainingTime[loan_id] -= 1
                else:
                    self.loanContractRemainingTime.pop(loan_id)

    # ========================================
    # Profits after corporate + carbon tax
    # ========================================
    def updateProfitsAfterTax(self, isC02Taxed=False):
        """Apply corporate tax and optionally carbon tax"""
        carbonTax = copy.copy(self.carbonTax) if isC02Taxed else 0
        if self.profits > 0:
            self.net_profit = (1 - self.p.taxRate) * self.profits - carbonTax
            self.updateTax(self.p.taxRate * self.profits + carbonTax)
        else:
            self.net_profit = 1 * self.profits - carbonTax
            self.updateTax(carbonTax)

    # ========================================
    # Bank response (called after bank computes obtainedCredit)
    # ========================================
    def adjustAccordingToBankState(self, obtainedCredit, eps=1e-8):
        """Check bank _calculate_running_loan for the function"""
        ## Credit obtained via bank
        # If a loan is granted, create a new vintage with a repayment horizon
        # if obtainedCredit > 0 and len(self.loanContractRemainingTime) == 0:
        if obtainedCredit > 0:
            self.loanObtained = obtainedCredit
            self.loanList.append(np.sum([self.getLoan()]))

            if self.useEnergy == "green":
                self.loanContractRemainingTime[len(self.loanList) - 1] = (
                    self.p.greenLoanRepayPeriod
                )
            else:
                self.loanContractRemainingTime[len(self.loanList) - 1] = (
                    self.p.brownLoanRepayPeriod
                )

        ## Adjust deposit and networth
        # Cash in, update NW and leverage
        self.updateDeposit(np.sum(obtainedCredit))  # adjust deposit if loan is granted
        self.depositList[-1] = self.deposit
        self.netWorth = self.depositList[-1] - sum(self.loanList)

        ## Update other financial variables
        self.DTE = sum(self.loanList) / (self.netWorth + 1e-8)
        self.iF = np.max([0, self.DTE / 100])
        if self.DTE < 0:
            self.defaultProb = 1
        else:
            self.defaultProb = 1 - np.exp(-self.p.defaultProbAlpha * self.DTE)
        # print("default Prob: ", self.defaultProb)
        self.iL = np.max([0, self.p.bankICB * (1 + self.defaultProb)])

        data = [self.soldProducts, self.wage_bill, len(self.workersList)]
        self.firmDataWriter.append(data)

    # ========================================
    # Capital stock & valuation update (depreciation + new investment)
    # ========================================
    def update_capital_value(self):
        """Update capital stock with depreciation and new investment"""
        # depreciation update
        self.set_capital(self.get_capital() * (1 - self.capital_depreciation))

        # reflect the purchase of new capital
        # Add newly purchased units (from capital_growth / transactions)
        self.update_capital_increase(np.sum(self.capital_growth))
        self.set_capital(self.get_capital() + self.get_capital_increase())

        # new value of capital
        # Revalue
        self.capital_value = self.get_capital() * self.capital_price

    # ========================================
    # Bankruptcy condition + re-entry logic (keeps sector composition)
    # ========================================
    def setBankruptcy(self):
        """Handle bankruptcy and firm re-entry"""
        if self.getBankrupt() == True or self.getNetWorth() < 0:
            self.model.bankrupt_count += 1
            self.non_loan = np.sum([self.loanList])

            # Rebirth policy: keep minimum counts of brown/green firms
            if "ConsumerGoods" in str(self):
                self.model.numCSFirmBankrupt += 1
                if (
                    np.sum(self.model.csfirm_agents.getUseEnergy() == "brown") > 2
                    and np.sum(self.model.csfirm_agents.getUseEnergy() == "green") > 2
                ):
                    if np.random.rand() < 0.5:
                        # Release all workers
                        for wid in self.workersList:
                            self.model.aliveConsumers.select(
                                self.model.aliveConsumers.getIdentity() == wid
                            ).receiveFiring()
                        self.useEnergyType("brown")
                        self.bankrupt_reset()
                    else:
                        for wid in self.workersList:
                            self.model.aliveConsumers.select(
                                self.model.aliveConsumers.getIdentity() == wid
                            ).receiveFiring()
                        self.useEnergyType("green")
                        self.bankrupt_reset()
                elif np.sum(self.model.csfirm_agents.getUseEnergy() == "brown") <= 2:
                    for wid in self.workersList:
                        self.model.aliveConsumers.select(
                            self.model.aliveConsumers.getIdentity() == wid
                        ).receiveFiring()
                    self.useEnergyType("brown")
                    self.bankrupt_reset()
                elif np.sum(self.model.csfirm_agents.getUseEnergy() == "green") <= 2:
                    for wid in self.workersList:
                        self.model.aliveConsumers.select(
                            self.model.aliveConsumers.getIdentity() == wid
                        ).receiveFiring()
                    self.useEnergyType("green")
                    self.bankrupt_reset()
            elif "CapitalGoods" in str(self):
                self.model.numCPFirmBankrupt += 1
                if (
                    np.sum(self.model.cpfirm_agents.getUseEnergy() == "brown") > 1
                    and np.sum(self.model.cpfirm_agents.getUseEnergy() == "green") > 1
                ):
                    if np.random.rand() < 0.5:
                        for wid in self.workersList:
                            self.model.aliveConsumers.select(
                                self.model.aliveConsumers.getIdentity() == wid
                            ).receiveFiring()
                        self.useEnergyType("brown")
                        self.bankrupt_reset()
                    else:
                        for wid in self.workersList:
                            self.model.aliveConsumers.select(
                                self.model.aliveConsumers.getIdentity() == wid
                            ).receiveFiring()
                        self.useEnergyType("green")
                        self.bankrupt_reset()
                elif np.sum(self.model.cpfirm_agents.getUseEnergy() == "brown") <= 1:
                    for wid in self.workersList:
                        self.model.aliveConsumers.select(
                            self.model.aliveConsumers.getIdentity() == wid
                        ).receiveFiring()
                    self.useEnergyType("brown")
                    self.bankrupt_reset()
                elif np.sum(self.model.cpfirm_agents.getUseEnergy() == "green") <= 1:
                    for wid in self.workersList:
                        self.model.aliveConsumers.select(
                            self.model.aliveConsumers.getIdentity() == wid
                        ).receiveFiring()
                    self.useEnergyType("green")
                    self.bankrupt_reset()

    # ========================================
    # Build wage bill (handles sick leave, lockdown, tax withholding)
    # ========================================
    def calculate_all_wages(self):
        """Calculate wages for all workers including sick leave and lockdown adjustments"""
        # Initialize the total wage bill and calculate the number of days in the current month
        self.wage_bill = 0
        self.wage_factor *= 1 + self.p.wageAdjustmentRate

        # Move invariant calculations OUT of the loop
        self.daysInMonth = days_in_month(
            int(str(self.model.today).split("-")[1]),
            int(str(self.model.today).split("-")[0]),
        )
        days = self.daysInMonth
        lockdown_days = len(self.lockdownList)
        unemployment_dole = self.p.unemploymentDole
        pandemic_transfer = self.p.pandemicWageTransfer

        # Optimization: Pre-compute worker set for O(1) lookup
        # self.workersList contains IDs of active workers
        workers_set = set(self.workersList)

        for aConsumer in self.consumersList:
            # Check if the consumer's identity is not in the workersList (i.e., not an active worker)
            # Use direct ID access avoid method call
            if aConsumer.id not in workers_set:
                continue

            # If the consumer's identity is not in the wages dictionary, add it and set the initial wage
            if aConsumer.id not in self.wages:
                self.wages[aConsumer.id] = aConsumer.getWage()

            wages = self.wages[aConsumer.id]
            # print("wage paid", wages)

            # Note: Logic assumes the commented out/overwritten wage calculation
            # from original code was ineffective for the payment, but used for accounting?
            # Preserving original flow logic but cleaning up

            sick_leaves = aConsumer.getSickLeaves()
            len_sick_leaves = len(sick_leaves)

            # Wage Setting
            if aConsumer.getEmploymentState():
                # The logic in original block lines 479-495 calculated a 'wage' that was
                # strictly OVERWRITTEN by line 498.
                # We retain the final calculation for payment.

                # Apply firm's wage factor and income tax withholding
                wage = (
                    wages
                    / (1 - self.p.incomeTaxRate)
                    * self.wage_factor
                    * (1 - self.p.incomeTaxRate)
                )

                # Update the consumer's wage
                aConsumer.setWage(wage)
                self.updateTax(wage * self.p.incomeTaxRate)
                # Update the total wage bill for this time step
                self.wage_bill += wage - (unemployment_dole / days) * len_sick_leaves

    # ========================================
    # Energy back-out from CES target (given L,K and planned Q)
    # ========================================
    def optimize_energy(self, labour, capital):
        """Calculate optimal energy input given labour, capital, and planned production"""
        energy = (1 / self.rho_energy) * (
            (
                self.planned_production**self.eta_production
                - (self.beta_labour * (self.rho_labour * labour) ** self.eta_production)
                - (self.beta_capital * (capital) ** self.eta_production)
            )
            / self.beta_energy
        ) ** (1 / self.eta_production)
        if (
            (
                self.planned_production**self.eta_production
                - (self.beta_labour * (self.rho_labour * labour) ** self.eta_production)
                - (self.beta_capital * (capital) ** self.eta_production)
            )
            / self.beta_energy
        ) < 0:
            energy = 0
        return energy

    # ========================================
    # Monthly payback, carbon tax pass-through, and risk metrics
    # ========================================
    def progressPayback(self, eps=1e-8):
        """Calculate loan payments and apply carbon tax surcharge to price"""
        # Amortization with simple annuity formula if there are active loans
        if len(self.loanContractRemainingTime) > 0:
            self.payback = (
                self.iL
                * np.sum([self.loanList])
                / (
                    1
                    - (1 + self.iL)
                    ** (list(self.loanContractRemainingTime.values())[0])
                )
            )
        else:
            self.payback = 0

        if self.p.verboseFlag:
            print("payback value", self.payback)

        # Carbon tax pass-through into price if brown and CT active
        brown_firm_coefficient = (
            self.brown_firm * self.p.climateZetaBrown * self.p.co2_price
        )
        carbon_tax = (
            brown_firm_coefficient
            * self.carbon_tax_state
            * self.get_actual_production()
        )
        self.setPrice(
            self.getPrice() + np.sum(carbon_tax / (self.get_actual_production()))
        )
        self.payLoan()

        ## Update other financial variables
        # Update risk metrics after payment
        self.DTE = sum(self.loanList) / (self.netWorth + eps)
        self.iF = np.max([0, self.DTE / 100])
        # print("default probability factor", self.DTE)
        if self.DTE < 0:
            self.defaultProb = 1
        else:
            self.defaultProb = 1 - np.exp(-self.p.defaultProbAlpha * self.DTE)
        self.iL = np.max([0, self.p.bankICB * (1 + self.defaultProb)])
        # print("default Prob: ", self.defaultProb, " ", self.DTE)

    # ========================================
    # Fire workers (used on shutdown/bankruptcy)
    # ========================================
    def fire(self):
        """Fire all workers (used on shutdown/bankruptcy)"""
        # function for firm to fire workers
        for (
            id
        ) in (
            self.workersList
        ):  # this line require an update mechanic to fire worker since the worker list shrink as we fire
            worker = self.model.aliveConsumers.select(
                self.model.aliveConsumers.getIdentity() == id
            )
            worker.receiveFiring()
            self.workersList.remove(id)

    # ========================================
    # Shocks & lockdown utilities
    # ========================================
    def setLockDown(self):
        """Activate lockdown state"""
        self.lockdown = True
        self.lockdownList.append(self.model.today)

    def unsetLockDown(self):
        """Deactivate lockdown state"""
        self.lockdown = False

    def resetLockDown(self):
        """Clear lockdown history"""
        self.lockdownList = []

    # ========================================
    # Fiscal transfer hook (e.g., targeted support)
    # ========================================
    def gov_transfer(self, value):
        """Receive government transfer"""
        self.updateDeposit(value)
        self.depositList[-1] = self.deposit
        self.netWorth += self.depositList[-1]
        self.model.government_agents.expenditure += value

    # ----------------------------------------
    # Small helpers/getters
    # ----------------------------------------
    def get_actual_production(self):
        return self.actual_production

    def getNumberOfLabours(self):
        return len(self.workersList)

    def update_actual_production(self, value):
        self.actual_production += value

    def getPrice(self):
        return self.price

    def getNetProfit(self):
        return self.net_profit

    def getIdentity(self):
        return self.id

    def getLoan(self):
        return self.loanObtained

    # ----------------------------------------
    # Demand aggregation helpers
    # ----------------------------------------
    def get_aggregate_demand(self):
        return self.aggregate_demand

    def set_aggregate_demand(self, value):
        self.aggregate_demand = value

    def update_aggregate_demand(self, value):
        self.aggregate_demand += value

    # ----------------------------------------
    # Simple setters/getters
    # ----------------------------------------
    def set_energy(self, value):
        self.set_energy = value

    def get_energy(self):
        return self.energy

    def set_capital(self, capital):
        self.capital = capital

    def get_capital(self):
        return self.capital

    def set_actual_production(self, actual_production):
        self.actual_production = actual_production

    def get_actual_production(self):
        return self.actual_production

    def set_capital_investment(self, capital_investment):
        self.capital_investment = capital_investment

    def get_capital_investment(self):
        return self.capital_investment

    def setSoldProducts(self, soldProducts):
        self.soldProducts = np.sum(soldProducts)

    def updateSoldProducts(self, soldProducts):
        self.soldProducts += np.sum(soldProducts)

    def getSoldProducts(self):
        return self.soldProducts

    def setPrice(self, price):
        self.price = price

    def getPrice(self):
        return self.price

    def useEnergyType(self, energyType):
        self.useEnergy = energyType

    def getUseEnergy(self):
        return self.useEnergy

    def getBankrupt(self):
        return self.bankrupt

    def getDeposit(self):
        return self.deposit

    def set_sale_record(self, sale):
        self.sale_record = sale

    def update_sale_record(self, value):
        self.sale_record += value

    def getNetWorth(self):
        return self.netWorth

    def getTax(self):
        return self.tax

    def setTax(self, tax):
        self.tax = tax

    def updateTax(self, value):
        self.tax += value

    def updateDeposit(self, value):
        self.deposit += value

    def append2DepositList(self, deposit):
        self.depositList.append(deposit)

    def get_average_production_cost(self):
        return self.average_production_cost

    def update_capital_increase(self, value):
        self.capital_increase = np.max([0, value])

    def get_capital_increase(self):
        return self.capital_increase

    def get_capital_demand(self):
        return self.capital_demand

    def set_capital_price(self, value):
        # firm will get this info when doing transaction, capital firm know their price
        # Provided by capital-goods firm during transaction
        self.capital_price = value

    def getOwnerIncome(self):
        return self.ownerIncome

    def reset_non_loan(self):
        self.non_loan = 0



#!/usr/bin/env python3
"""
Tests for individual model components in CliMaPan-Lab.
"""

import os
import sys
import unittest

import numpy as np

# Add the climapan_lab package to the path
sys.path.insert(0, os.path.join(os.path.dirname(__file__), ".."))

try:
    from climapan_lab.base_params import economic_params
    from climapan_lab.src.models import EconModel
    from climapan_lab.src.params import parameters

    IMPORTS_AVAILABLE = True
except ImportError as e:
    IMPORTS_AVAILABLE = False
    IMPORT_ERROR = str(e)


class TestModelComponents(unittest.TestCase):
    """Test individual components of the economic model."""

    def setUp(self):
        """Set up test fixtures."""
        if not IMPORTS_AVAILABLE:
            self.skipTest(f"Required imports not available: {IMPORT_ERROR}")

        self.params = economic_params.copy()
        # Use very small numbers for fast testing
        self.params.update(
            {
                "c_agents": 10,
                "capitalists": 3,
                "csf_agents": 2,  # Need at least 2 for energy type diversity
                "cpf_agents": 2,  # Need at least 2 for energy type diversity
                "green_energy_owners": 1,
                "brown_energy_owners": 1,
                "b_agents": 1,
                "g_agents": 1,
                "steps": 3,
                "verboseFlag": False,
                "climateModuleFlag": False,  # Disable for fast testing
                "covid_settings": None,  # Disable COVID for testing
            }
        )

    def test_model_initialization(self):
        """Test that model initializes with correct agent counts."""
        model = EconModel(self.params)
        model.setup()

        # Check agent counts
        self.assertEqual(len(model.consumer_agents), self.params["c_agents"])
        self.assertEqual(len(model.csfirm_agents), self.params["csf_agents"])
        self.assertEqual(len(model.cpfirm_agents), self.params["cpf_agents"])

        # Check that model has required attributes
        self.assertTrue(hasattr(model, "t"))
        self.assertTrue(hasattr(model, "p"))
        self.assertEqual(model.t, 0)

    def test_agent_attributes(self):
        """Test that agents have required attributes."""
        model = EconModel(self.params)
        model.setup()

        # Test consumer attributes
        if len(model.consumer_agents) > 0:
            consumer = model.consumer_agents[0]
            required_attrs = ["deposit", "employed", "consumerType"]
            for attr in required_attrs:
                self.assertTrue(
                    hasattr(consumer, attr), f"Consumer missing attribute: {attr}"
                )

        # Test firm attributes
        if len(model.csfirm_agents) > 0:
            firm = model.csfirm_agents[0]
            required_attrs = ["netWorth", "price", "actual_production"]
            for attr in required_attrs:
                self.assertTrue(hasattr(firm, attr), f"Firm missing attribute: {attr}")

    def test_climate_module_toggle(self):
        """Test that climate module can be enabled/disabled."""
        # Test with climate disabled
        params_no_climate = self.params.copy()
        params_no_climate["climateModuleFlag"] = False
        model_no_climate = EconModel(params_no_climate)

        # Test with climate enabled
        params_with_climate = self.params.copy()
        params_with_climate["climateModuleFlag"] = True
        model_with_climate = EconModel(params_with_climate)

        # Both should initialize successfully
        self.assertIsInstance(model_no_climate, EconModel)
        self.assertIsInstance(model_with_climate, EconModel)

    def test_covid_scenarios(self):
        """Test different COVID scenarios."""
        covid_settings = [None, "BAU", "DIST", "LOCK", "VAX"]

        for covid_setting in covid_settings:
            with self.subTest(covid_setting=covid_setting):
                params = self.params.copy()
                params["covid_settings"] = covid_setting

                # Should initialize without errors
                model = EconModel(params)
                self.assertIsInstance(model, EconModel)

    def test_economic_scenarios(self):
        """Test different economic scenarios."""
        economic_settings = ["BAU", "CT", "CTRa", "CTRb", "CTRc", "CTRd"]

        for setting in economic_settings:
            with self.subTest(setting=setting):
                params = self.params.copy()
                params["settings"] = setting

                # Should initialize without errors
                model = EconModel(params)
                self.assertIsInstance(model, EconModel)

    def test_parameter_validation(self):
        """Test that parameters are correctly assigned to models."""
        # Test that parameters are properly accessible
        model = EconModel(self.params)
        self.assertTrue(hasattr(model, "p"))

        # Test some key parameters
        for key, value in self.params.items():
            if hasattr(model.p, key):
                self.assertEqual(getattr(model.p, key), value)

    def test_model_step_progression(self):
        """Test that model time progresses correctly."""
        model = EconModel(self.params)
        model.setup()
        initial_time = model.t

        # Take a few steps and verify they execute without error
        for i in range(3):
            try:
                model.step()
                step_completed = True
            except Exception:
                step_completed = False
            self.assertTrue(step_completed, f"Step {i+1} should complete without error")

    def test_agent_state_changes(self):
        """Test that agent states change during simulation."""
        model = EconModel(self.params)
        model.setup()

        # Record initial states
        if len(model.consumer_agents) > 0:
            initial_consumer_budget = model.consumer_agents[0].deposit

        # Run a few steps
        for _ in range(2):
            model.step()

        # States should potentially change (though we can't guarantee specific changes)
        # At minimum, the model should still be in a valid state
        if len(model.consumer_agents) > 0:
            self.assertIsInstance(model.consumer_agents[0].deposit, (int, float))

    def test_data_collection(self):
        """Test that model collects data properly."""
        model = EconModel(self.params)

        # Run model briefly
        results = model.run()

        # Check that data collection works (ambr returns dict with 'model' DataFrame)
        # Legacy AgentPy checked model.output.variables
        # We now check the returned results
        if isinstance(results, dict):
            self.assertIn("model", results)
            self.assertFalse(results["model"].is_empty())
        else:
            # If wrapper is used or AgentPy compat matches
            self.assertTrue(hasattr(results, "variables"))


class TestParameterStructure(unittest.TestCase):
    """Test parameter structure and consistency."""

    def test_parameter_types(self):
        """Test that parameters have correct types."""
        if not IMPORTS_AVAILABLE:
            self.skipTest(f"Required imports not available: {IMPORT_ERROR}")

        params = economic_params

        # Test integer parameters
        int_params = ["c_agents", "steps", "capitalists", "csf_agents"]
        for param in int_params:
            if param in params:
                self.assertIsInstance(params[param], int, f"{param} should be int")
                self.assertGreater(params[param], 0, f"{param} should be positive")

        # Test float parameters
        float_params = ["taxRate", "co2_tax", "depreciationRate"]
        for param in float_params:
            if param in params:
                self.assertIsInstance(
                    params[param], (int, float), f"{param} should be numeric"
                )

        # Test boolean parameters
        bool_params = ["verboseFlag", "energySectorFlag", "climateModuleFlag"]
        for param in bool_params:
            if param in params:
                self.assertIsInstance(params[param], bool, f"{param} should be boolean")

    def test_required_parameters(self):
        """Test that all required parameters are present."""
        if not IMPORTS_AVAILABLE:
            self.skipTest(f"Required imports not available: {IMPORT_ERROR}")

        params = economic_params

        required_params = [
            "c_agents",
            "steps",
            "taxRate",
            "unemploymentDole",
            "subsistenceLevelOfConsumption",
            "settings",
        ]

        for param in required_params:
            self.assertIn(param, params, f"Required parameter {param} missing")

    def test_parameter_ranges(self):
        """Test that parameters are within reasonable ranges."""
        if not IMPORTS_AVAILABLE:
            self.skipTest(f"Required imports not available: {IMPORT_ERROR}")

        params = economic_params

        # Test that rates are between 0 and 1
        rate_params = ["taxRate", "incomeTaxRate", "reserve_ratio"]
        for param in rate_params:
            if param in params:
                self.assertGreaterEqual(params[param], 0, f"{param} should be >= 0")
                self.assertLessEqual(params[param], 1, f"{param} should be <= 1")

        # Test that agent counts are positive
        agent_params = ["c_agents", "capitalists", "csf_agents", "cpf_agents"]
        for param in agent_params:
            if param in params:
                self.assertGreater(params[param], 0, f"{param} should be positive")


class TestErrorHandling(unittest.TestCase):
    """Test error handling and edge cases."""

    def test_empty_parameters(self):
        """Test behavior with empty or missing parameters."""
        if not IMPORTS_AVAILABLE:
            self.skipTest(f"Required imports not available: {IMPORT_ERROR}")

        # Test that model can be created (AgentPy is lenient with parameters)
        try:
            model = EconModel({})
            model_created = True
        except Exception:
            model_created = False
        # Either behavior is acceptable - just test that it's consistent
        self.assertIsInstance(model_created, bool)

    def test_string_numeric_parameters(self):
        """Test that models handle parameter types appropriately."""
        if not IMPORTS_AVAILABLE:
            self.skipTest(f"Required imports not available: {IMPORT_ERROR}")

        params = economic_params.copy()
        params["c_agents"] = "not_a_number"

        # Test that model can be created (validation happens during setup/run)
        try:
            model = EconModel(params)
            model_created = True
        except Exception:
            model_created = False
        # Either behavior is acceptable
        self.assertIsInstance(model_created, bool)

    def test_negative_steps(self):
        """Test that models handle step parameters appropriately."""
        if not IMPORTS_AVAILABLE:
            self.skipTest(f"Required imports not available: {IMPORT_ERROR}")

        params = economic_params.copy()
        params["steps"] = -1

        # Test that model can be created (AgentPy handles this)
        try:
            model = EconModel(params)
            model_created = True
        except Exception:
            model_created = False
        # Either behavior is acceptable
        self.assertIsInstance(model_created, bool)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
"""
Basic simulation example for CliMaPan-Lab.

This example demonstrates how to:
1. Set up and run a basic economic simulation
2. Configure different scenarios (BAU, Climate Tax, COVID policies)
3. Save and analyze results
4. Create basic visualizations

Usage:
    python basic_simulation.py [--scenario BAU|CT|CTR] [--covid BAU|DIST|LOCK|VAX]
"""

import argparse
import os
from datetime import datetime

import matplotlib.pyplot as plt
import numpy as np

from ..base_params import economic_params

# Import the economic model and parameters
from ..model import EconModel
from ..run_sim import single_run


def setup_basic_scenario(
    scenario="BAU", covid_policy="BAU", steps=120, agents_scale=1.0
):
    """
    Set up a basic scenario with configurable parameters.

    Args:
        scenario: Economic scenario ("BAU", "CT", "CTR")
        covid_policy: COVID policy ("BAU", "DIST", "LOCK", "VAX")
        steps: Number of simulation steps (months)
        agents_scale: Scale factor for agent numbers (for testing vs full runs)

    Returns:
        dict: Parameters for the simulation
    """
    params = economic_params.copy()

    # Scale agent numbers for testing vs production
    params.update(
        {
            "c_agents": int(params["c_agents"] * agents_scale),
            "capitalists": int(params["capitalists"] * agents_scale),
            "csf_agents": max(1, int(params["csf_agents"] * agents_scale)),
            "cpf_agents": max(1, int(params["cpf_agents"] * agents_scale)),
            "green_energy_owners": max(
                1, int(params["green_energy_owners"] * agents_scale)
            ),
            "brown_energy_owners": max(
                1, int(params["brown_energy_owners"] * agents_scale)
            ),
            "steps": steps,
            "settings": scenario,
            "covid_settings": covid_policy,
            "verboseFlag": False,
        }
    )

    # Scenario-specific adjustments
    if scenario == "CT":  # Carbon Tax
        params["co2_tax"] = 0.05
        params["co2_price"] = 50
    elif scenario.startswith("CTR"):  # Carbon Tax with Redistribution
        params["co2_tax"] = 0.05
        params["co2_price"] = 50
        params["lumpSum"] = 500000  # Redistribution amount

    # COVID policy adjustments
    if covid_policy == "DIST":  # Social Distancing
        params["num_contacts_community"] = 10  # Reduced contacts
        params["num_contacts_firms"] = 10
    elif covid_policy == "LOCK":  # Lockdown
        params["num_contacts_community"] = 5
        params["num_contacts_firms"] = 5
        params["sick_reduction"] = 20  # Higher consumption reduction when sick
    elif covid_policy == "VAX":  # Vaccination
        params["p_vax"] = 0.8  # Vaccination rate

    return params


def run_basic_simulation(params, output_dir="results"):
    """
    Run a basic simulation with the given parameters.

    Args:
        params: Simulation parameters
        output_dir: Directory to save results

    Returns:
        tuple: (model_results, save_folder)
    """
    print(
        f"Running simulation: {params['settings']} scenario with {params['covid_settings']} COVID policy"
    )
    print(
        f"Agents: {params['c_agents']} consumers, {params['csf_agents']} consumer firms, {params['cpf_agents']} capital firms"
    )
    print(f"Steps: {params['steps']} months")

    # Ensure output directory exists
    os.makedirs(output_dir, exist_ok=True)

    # Run the simulation
    results = single_run(params, parent_folder=output_dir, make_stats=True)

    return results


def create_basic_plots(results, save_folder):
    """
    Create basic plots from simulation results.

    Args:
        results: Simulation results
        save_folder: Directory to save plots
    """
    if not hasattr(results, "variables") or not hasattr(results.variables, "EconModel"):
        print("Warning: Results don't contain expected variables for plotting")
        return

    data = results.variables.EconModel

    # Create figures directory
    plots_dir = os.path.join(save_folder, "plots")
    os.makedirs(plots_dir, exist_ok=True)

    # Plot 1: GDP over time
    plt.figure(figsize=(10, 6))
    if "GDP" in data:
        gdp_data = data["GDP"].values
        if len(gdp_data) > 0 and hasattr(gdp_data[0], "__len__"):
            gdp_series = [x for x in gdp_data if x is not None]
            if gdp_series:
                plt.plot(gdp_series)
                plt.title("GDP Over Time")
                plt.xlabel("Time (months)")
                plt.ylabel("GDP")
                plt.grid(True)
                plt.savefig(os.path.join(plots_dir, "gdp_over_time.png"))
                plt.close()

    # Plot 2: Unemployment rate
    plt.figure(figsize=(10, 6))
    if "UnemploymentRate" in data:
        unemployment_data = data["UnemploymentRate"].values
        unemployment_series = [
            x for x in unemployment_data if x is not None and str(x) != "nan"
        ]
        if unemployment_series:
            plt.plot(unemployment_series)
            plt.title("Unemployment Rate Over Time")
            plt.xlabel("Time (months)")
            plt.ylabel("Unemployment Rate")
            plt.grid(True)
            plt.savefig(os.path.join(plots_dir, "unemployment_over_time.png"))
            plt.close()

    # Plot 3: Climate variables (if available)
    if "ClimateTemperature" in data:
        plt.figure(figsize=(10, 6))
        temp_data = data["ClimateTemperature"].values
        temp_series = [x for x in temp_data if x is not None]
        if temp_series:
            plt.plot(temp_series)
            plt.title("Temperature Anomaly Over Time")
            plt.xlabel("Time (months)")
            plt.ylabel("Temperature Anomaly (°C)")
            plt.grid(True)
            plt.savefig(os.path.join(plots_dir, "temperature_over_time.png"))
            plt.close()

    print(f"Basic plots saved to {plots_dir}")


def analyze_results(results, save_folder):
    """
    Perform basic analysis of simulation results.

    Args:
        results: Simulation results
        save_folder: Directory to save analysis
    """
    if not hasattr(results, "variables") or not hasattr(results.variables, "EconModel"):
        print("Warning: Results don't contain expected variables for analysis")
        return

    data = results.variables.EconModel
    analysis_file = os.path.join(save_folder, "basic_analysis.txt")

    with open(analysis_file, "w") as f:
        f.write("Basic Simulation Analysis\n")
        f.write("=" * 30 + "\n\n")

        # Analyze key economic indicators
        for var_name in ["GDP", "UnemploymentRate", "InflationRate", "Consumption"]:
            if var_name in data:
                var_data = data[var_name].values
                clean_data = [x for x in var_data if x is not None and str(x) != "nan"]

                if clean_data:
                    f.write(f"{var_name}:\n")
                    f.write(f"  Mean: {np.mean(clean_data):.4f}\n")
                    f.write(f"  Std:  {np.std(clean_data):.4f}\n")
                    f.write(f"  Min:  {np.min(clean_data):.4f}\n")
                    f.write(f"  Max:  {np.max(clean_data):.4f}\n\n")

        # Analyze climate variables if available
        for var_name in ["ClimateTemperature", "ClimateC02Concentration"]:
            if var_name in data:
                var_data = data[var_name].values
                clean_data = [x for x in var_data if x is not None]

                if clean_data:
                    f.write(f"{var_name}:\n")
                    f.write(f"  Final value: {clean_data[-1]:.4f}\n")
                    f.write(f"  Change: {clean_data[-1] - clean_data[0]:.4f}\n\n")

    print(f"Basic analysis saved to {analysis_file}")


def main():
    """Main function to run the basic simulation example."""
    parser = argparse.ArgumentParser(description="Run a basic CliMaPan-Lab simulation")
    parser.add_argument(
        "--scenario",
        choices=["BAU", "CT", "CTR"],
        default="BAU",
        help="Economic scenario to run",
    )
    parser.add_argument(
        "--covid",
        choices=["BAU", "DIST", "LOCK", "VAX"],
        default="BAU",
        help="COVID policy to apply",
    )
    parser.add_argument(
        "--steps", type=int, default=120, help="Number of simulation steps (months)"
    )
    parser.add_argument(
        "--scale",
        type=float,
        default=0.1,
        help="Agent scale factor (0.1 for quick testing, 1.0 for full simulation)",
    )
    parser.add_argument(
        "--output", default="results", help="Output directory for results"
    )
    parser.add_argument("--plots", action="store_true", help="Generate basic plots")
    parser.add_argument(
        "--analysis", action="store_true", help="Perform basic analysis"
    )

    args = parser.parse_args()

    # Set up the scenario
    params = setup_basic_scenario(
        scenario=args.scenario,
        covid_policy=args.covid,
        steps=args.steps,
        agents_scale=args.scale,
    )

    # Run the simulation
    try:
        results = run_basic_simulation(params, args.output)
        print("Simulation completed successfully!")

        # Find the save folder (created by single_run)
        timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
        save_folder = None

        # Look for the results folder
        for item in os.listdir(args.output):
            if item.startswith(f"results_{args.scenario}_{args.covid}"):
                save_folder = os.path.join(args.output, item)
                break

        if save_folder and os.path.exists(save_folder):
            print(f"Results saved to: {save_folder}")

            # Generate plots if requested
            if args.plots:
                create_basic_plots(results, save_folder)

            # Perform analysis if requested
            if args.analysis:
                analyze_results(results, save_folder)
        else:
            print("Warning: Could not locate results folder for post-processing")

    except Exception as e:
        print(f"Simulation failed with error: {e}")
        raise


if __name__ == "__main__":
    main()
